//----------------------------------------------------------------------------------------------
// BSP : Seeed nRF52 Borads 1.1.1
// Board : Seeed nRF52 Borads / Seeed XIAO nRF52840
//----------------------------------------------------------------------------------------------

// 2023/06/07
// Central sketch : nRF52_XIAO_LightSleepBatteryTest_peripheral.ino

#include <Arduino.h>
#include <Adafruit_TinyUSB.h>  // For Serial.print()
#include <Wire.h>
#include <SPI.h>
#include <SD.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>
#include "Fonts/FreeSans9pt7b.h"
#include "Fonts/FreeSerif9pt7b.h"
#include <bluefruit.h>  // \Arduino15\packages\Seeeduino\hardware\nrf52\1.1.1\libraries\Bluefruit52Lib\src

#define HICHG       	 22   	// charging current　HIGH(Open):50mA LOW:100mA
#define CHG         	 23   	// charge indicator Discharge:HIGH,LED OFF　Charge:LOW,LED ON
#define VBAT_ENABLE 	 14   	// battery voltage readable　HIGH(Open):disable　LOW:enable
#define VBAT_READ   	 32   	// battery voltage ADC input
#define SD_CS       	 D2     // SD chis select pin
#define LOOP_TIME    	 1		  // Loop time [sec]

#define dataNum 8                 	// Vbatt 4, mS 4 (2 data 8 byte)
union unionData {                 	// Union for data type conversion
  float     dataBuffFl[dataNum/4];
  uint32_t  dataBuff32[dataNum/4];
  uint8_t   dataBuff8[dataNum];
};
union unionData ud;
const char* versionNumber = "0.99"; // Central sketch version number
uint32_t mS;                        // Measurement time [mS]
float minute;						            // Measurement time [min.]
bool notifyFlag = false;            // Notify call back flag
float VbattP;                       // Peripheral Battery voltage [V]
float VbattC;                       // Central Battery voltage [V]

// Instances of the class
Adafruit_SSD1306 display(128, 64, &Wire, -1);
File logFile; 

// Custum Service and Characteristic
// 53c40000-8682-4dd1-be0c-40588193b485 for example
#define vbattService_UUID(val) (const uint8_t[]) { \
    0x85, 0xB4, 0x93, 0x81, 0x58, 0x40, 0x0C, 0xBE, 0xD1, 0x4D, \
    (uint8_t)(val & 0xff), (uint8_t)(val >> 8), \
    0x00, 0x00, 0xC4, 0x53 }

BLEClientService        vbattService         (vbattService_UUID(0x0000));
BLEClientCharacteristic dataCharacteristic   (vbattService_UUID(0x0030));

void setup()
{
  // initialization of Serial port 
  Serial.begin(115200);
//  while (!Serial);
  delay(2000);

  // Pin initialization
  pinMode(VBAT_READ, INPUT); 
  pinMode(HICHG, OUTPUT);
  pinMode(VBAT_ENABLE, OUTPUT);   
  pinMode(LED_RED, OUTPUT);
  pinMode(LED_GREEN, OUTPUT);
  pinMode(LED_BLUE, OUTPUT); 

  digitalWrite(HICHG, LOW);         // Charge current 100mA
  digitalWrite(VBAT_ENABLE, LOW);   // Battery voltage read enable   
  digitalWrite(LED_RED, HIGH);
  digitalWrite(LED_GREEN, HIGH);
  digitalWrite(LED_BLUE, HIGH);

  // ADC initialization　See wireing_analog_nRF52.c:73
  analogReference(AR_DEFAULT);      // Default 0.6V*6=3.6V
  analogReadResolution(12);         // wireing_analog_nRF52.c:39
  analogSampleTime(10);             // default 3uS　3,5,10,15,20,40

  // Initialization of display SSD1306
  Wire.begin();
  if (!display.begin(SSD1306_SWITCHCAPVCC, 0x3C))
  {
    Serial.println("Initialization failure of SSD1306");
    while(1) {
      digitalWrite(LED_BLUE, LOW);
    }
  }
  display.clearDisplay();
  display.setFont(&FreeSans9pt7b);
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(0, 15);
  display.println("Light Sleep");
  display.setCursor(0, 31);
  display.println("BLE Central");
  display.setCursor(0, 47);
  display.print("Ver   : ");
  display.println(versionNumber);
  display.display();
  delay(5000);
  display.clearDisplay();

  // Initialization of SD
  if(!SD.begin(32000000UL, SD_CS)) {    // default 4MHz  
    Serial.println("SD card initialization failed!");
    display.clearDisplay();
    display.setCursor(0, 15);
    display.println("SD init failed");
    display.display();
    while(1) {
      digitalWrite(LED_BLUE, LOW);
    }
  } 

  // Log file open
  logFile = SD.open("log.csv", FILE_WRITE);
  if(!logFile) { 
    Serial.println("log.csv can't open");
    display.clearDisplay();
    display.setCursor(0, 15);
    display.println("log can't open");
    display.display();   
  }

  // Initialization of Bruefruit
  Bluefruit.configCentralBandwidth(BANDWIDTH_MAX);  
  Bluefruit.begin(1, 1);
  Bluefruit.setName("XIAO BLE Central");

  // Custom Service Settings
  vbattService.begin();
  dataCharacteristic.setNotifyCallback(data_notify_callback);
  dataCharacteristic.begin();

  // Blue LED blinking interval setting
  Bluefruit.setConnLedInterval(100);

  // Callbacks
  Bluefruit.Central.setDisconnectCallback(disconnect_callback);
  Bluefruit.Central.setConnectCallback(connect_callback);

  // Scanner settings
  Bluefruit.Scanner.setRxCallback(scan_callback);
  Bluefruit.Scanner.restartOnDisconnect(true);
  Bluefruit.Scanner.setIntervalMS(100, 50);     // interval:100mS  window:50mS  
  Bluefruit.Scanner.filterUuid(vbattService.uuid);
  Bluefruit.Scanner.useActiveScan(false);
  Bluefruit.Scanner.start(0);                   // Continuous Scanning

  Serial.println("End of initialization");   
}

//******************************************************************************************
void loop()
{
  uint32_t timestamp = millis();

  // Wait for response from peripheral
  Serial.println("Connecting to Peripheral ....");
  while (!Bluefruit.connected())
  {
    Serial.print(".");
    delay(1000);
  }
  Serial.println("");
  Serial.println("Connected to Peripheral");

  digitalWrite(LED_GREEN, LOW);

  if(notifyFlag == true)  // set true in data_notify_callback()
  {

    // Central battery voltage averaging
    int Vadc = 0;
    for(int i = 0; i < 16; i++)
    {
      Vadc = Vadc + analogRead(VBAT_READ);
    }
    Vadc = Vadc / 16; 
    VbattC = 2.961 * 3.6 * Vadc / 4096;   // Vref=3.6 ratio=2.961
    
    // Restore incoming data     
    VbattP = ud.dataBuffFl[0];
    mS = ud.dataBuff32[1];
    minute = mS / (1000.0 * 60.0);

    // Data display and save
    Serial.print(VbattP, 3);
    Serial.print(" ");
    Serial.println(minute, 2);

    logFile.print(minute);
    logFile.print(",");
    logFile.println(VbattP);
    logFile.flush();  

    display.clearDisplay();
    display.setCursor(0, 15);
    display.print("VbattP "); display.print(VbattP, 3);
    display.setCursor(0, 31);
    display.print("mS "); display.print(mS);
    display.setCursor(0, 47);
    display.print("minute "); display.print(minute);
    display.setCursor(0,63);
    display.print("VbattC "); display.print(VbattC, 3);
    display.display();

    notifyFlag = false;
  } 
  digitalWrite(LED_GREEN, HIGH);

  while(millis() - timestamp < 1000 * LOOP_TIME);
}

//*****************************************************************************************************
// Callback invoked when scanner pick up an advertising data
void scan_callback(ble_gap_evt_adv_report_t* report)
{
  Bluefruit.Central.connect(report);
}

//****************************************************************************************************:
// Callback invoked when an connection is established
void connect_callback(uint16_t conn_handle)
{
  if ( !vbattService.discover(conn_handle) )
  {
    Serial.println("【connect_callback】 Service not find disconnect!");
    Bluefruit.disconnect(conn_handle);
    return;
  }
  Serial.println("【connect_callback】 Find Service");
  
  if ( !dataCharacteristic.discover() )
  {
    Serial.println("【connect_callback】 Characteristic not find disconnect!");
    Bluefruit.disconnect(conn_handle);
    return;
  }
  Serial.println("【connect_callback】 Find Characteristic");
  
  if ( !dataCharacteristic.enableNotify() )
  {
    Serial.println("【connect_callback】 Couldn't enable notify for data");
    Bluefruit.disconnect(conn_handle);
    return;
  }
  Serial.println("【connect_callback】 enable notify");
}

//**********************************************************************************************************
// Callback invoked when an connection is broken
void disconnect_callback(uint16_t conn_handle, uint8_t reason)
{
  (void) conn_handle;
  (void) reason;

  Serial.print("【disconnect_callback】 Disconnected, reason = 0x"); Serial.println(reason, HEX);
}

//*********************************************************************************************************
// Callback invoked when a notifyication is received
void data_notify_callback(BLEClientCharacteristic* chr, uint8_t* data, uint16_t len)
{
  Serial.println("【notify_callback】 notify_callback, len = "); Serial.println(len);
  
  digitalWrite(LED_RED, LOW);

  // Data received
  memcpy(&ud.dataBuff8, data, len);
  
  delayMicroseconds(10000);   // 10mS for LED
  digitalWrite(LED_RED, HIGH);

  // notifyFlag set
  notifyFlag = true;
}
